#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/extensions/XInput.h>
#include <X11/Xutil.h>
#include <jpeglib.h>

int main(void)
{
   Display *display = NULL;
   int i = 0, devnum = 0, buttonpress;
   XDeviceInfo *devinfo = NULL;
   XDevice **device = NULL;
   XEventClass evlist;
   XEvent event;
   Window root, child;
   int rootX, rootY, childX, childY;
   unsigned int mask;
   int h = 10;

   int screen;
   Window window;
   XWindowAttributes wininfo;
   XImage *image = NULL;

   unsigned long pixel;
   int x, y;
   char *buffer;
   struct jpeg_compress_struct cinfo;
   struct jpeg_error_mgr jerr;
   JSAMPROW row;
   unsigned char *outbuf = NULL;
   unsigned long int outlen = 0;

   do {
      if((display = XOpenDisplay(NULL)) == NULL) break;
      if(!(devinfo = XListInputDevices(display, &devnum)) || !devnum) break;

      if(!(device = malloc(devnum * sizeof(XDevice *)))) break;

      for(i = 0; i < devnum; i++) {
         if(devinfo[i].use == IsXExtensionPointer) {
            if(!(device[i] = XOpenDevice(display, devinfo[i].id))) continue;
            DeviceButtonPress(device[i], buttonpress, evlist);
            XSelectExtensionEvent(display, DefaultRootWindow(display), &evlist, 1);
         } else {
            device[i] = NULL;
         }
      }

      while(1) {
         if(XNextEvent(display, &event)) continue;
         if(event.type != buttonpress) continue;
         XQueryPointer(display, DefaultRootWindow(display), &root, &child, &rootX, &rootY, &childX, &childY, &mask); 
         printf("x=%d, y=%d\n", rootX, rootY);

         screen = DefaultScreen(display);
         window = RootWindow(display, screen);
         if(!XGetWindowAttributes(display, window, &wininfo)) break;
         if(!(image = XGetImage(display, window, ((rootX > 20) ? rootX - 20 : 0), ((rootY > 20) ? rootY - 20 : 0), ((rootX + 20 < wininfo.width) ? 40 : wininfo.width - rootX), ((rootY + 20 < wininfo.height) ? 40 : wininfo.height - rootY), AllPlanes, ZPixmap))) break;
         if(image->depth != 24) break;

         buffer = malloc(3 * image->width * image->height);
         for(y = 0; y < image->height; y++) {
            for(x = 0; x < image->width; x++) {
               pixel = XGetPixel(image, x, y);
               buffer[y * image->width * 3 + x * 3 + 0] = (char)(pixel>>16);
               buffer[y * image->width * 3 + x * 3 + 1] = (char)((pixel>>8) & 0xff);
               buffer[y * image->width * 3 + x * 3 + 2] = (char)(pixel & 0xff);
            }
         }

         cinfo.err = jpeg_std_error(&jerr);
         jpeg_create_compress(&cinfo);
         jpeg_mem_dest(&cinfo, &outbuf, &outlen);

         cinfo.image_width = image->width;
         cinfo.image_height = image->height;
         cinfo.input_components = 3;
         cinfo.in_color_space = JCS_RGB;

         jpeg_set_defaults(&cinfo);
         jpeg_set_quality(&cinfo, 75, TRUE);
         jpeg_start_compress(&cinfo, TRUE);

         while(cinfo.next_scanline < cinfo.image_height) {
            row = (JSAMPROW)&buffer[cinfo.next_scanline * 3 * image->width];
            jpeg_write_scanlines(&cinfo, &row, 1);
         }
         jpeg_finish_compress(&cinfo);
         jpeg_destroy_compress(&cinfo);

         write(h++, outbuf, outlen);
      }
   } while(0);

   for(i = 0; i < devnum; i++) if(device[i]) XCloseDevice(display, device[i]); 
   if(device) free(device);
   if(devinfo) XFreeDeviceList(devinfo);
   if(display) XCloseDisplay(display);

   return 0;
}
