#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <linux/videodev2.h>
#include <libv4l2.h>
#include <jpeglib.h>

static void xioctl(int fh, int request, void *arg)
{
   int r;

   do {
      r = v4l2_ioctl(fh, request, arg);
   } while (r == -1 && ((errno == EINTR) || (errno == EAGAIN)));

   if (r == -1) {
      fprintf(stderr, "error %d, %s\n", errno, strerror(errno));
      exit(EXIT_FAILURE);
   }
}

int main(void)
{
   struct v4l2_format fmt;
   struct v4l2_buffer buf;
   struct v4l2_requestbuffers req;
   enum v4l2_buf_type type;
   struct {
      void   *start;
      size_t length;
   } *buffers;

   struct jpeg_compress_struct cinfo;
   struct jpeg_error_mgr jerr;
   JSAMPROW row;
   unsigned char *outbuf = NULL;
   unsigned long int outlen = 0;

   fd_set fds;
   struct timeval tv;
   int i, fd = -1;
   FILE *fout;

   if((fd = v4l2_open("/dev/video0", O_RDWR | O_NONBLOCK, 0)) < 0) {
      perror("Cannot open device");
      exit(EXIT_FAILURE);
   }

   memset(&fmt, 0x00, sizeof(fmt));
   fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
   fmt.fmt.pix.width = 640;
   fmt.fmt.pix.height = 480;
   fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_RGB24;
   fmt.fmt.pix.field = V4L2_FIELD_INTERLACED;
   xioctl(fd, VIDIOC_S_FMT, &fmt);
   if(fmt.fmt.pix.pixelformat != V4L2_PIX_FMT_RGB24) {
      printf("Libv4l didn't accept RGB24 format. Can't proceed.\n");
      exit(EXIT_FAILURE);
   }

   memset(&req, 0x00, sizeof(req));
   req.count = 2;
   req.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
   req.memory = V4L2_MEMORY_MMAP;
   xioctl(fd, VIDIOC_REQBUFS, &req);

   buffers = calloc(req.count, sizeof(*buffers));
   for(i = 0; i < req.count; i++) {
      memset(&buf, 0x00, sizeof(buf));

      buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
      buf.memory = V4L2_MEMORY_MMAP;
      buf.index = i;

      xioctl(fd, VIDIOC_QUERYBUF, &buf);

      buffers[i].length = buf.length;
      buffers[i].start = v4l2_mmap(NULL, buf.length, PROT_READ | PROT_WRITE, MAP_SHARED, fd, buf.m.offset);

      if(buffers[i].start == MAP_FAILED) {
         perror("mmap");
         exit(EXIT_FAILURE);
      }
   }

   for(i = 0; i < req.count; i++) {
      memset(&buf, 0x00, sizeof(buf));
      buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
      buf.memory = V4L2_MEMORY_MMAP;
      buf.index = i;
      xioctl(fd, VIDIOC_QBUF, &buf);
   }

   type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
   xioctl(fd, VIDIOC_STREAMON, &type);

   FD_ZERO(&fds);
   FD_SET(fd, &fds);
   tv.tv_sec = 5;
   tv.tv_usec = 0;
   if(select(fd + 1, &fds, NULL, NULL, &tv) <= 0) {
      perror("select");
      exit(EXIT_FAILURE);
   }
   memset(&buf, 0x00, sizeof(buf));
   buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
   buf.memory = V4L2_MEMORY_MMAP;
   xioctl(fd, VIDIOC_DQBUF, &buf);

   cinfo.err = jpeg_std_error(&jerr);
   jpeg_create_compress(&cinfo);
   jpeg_mem_dest(&cinfo, &outbuf, &outlen);

   cinfo.image_width = fmt.fmt.pix.width;
   cinfo.image_height = fmt.fmt.pix.height;
   cinfo.input_components = 3;
   cinfo.in_color_space = JCS_RGB;

   jpeg_set_defaults(&cinfo);
   jpeg_set_quality(&cinfo, 75, TRUE);
   jpeg_start_compress(&cinfo, TRUE);

   while(cinfo.next_scanline < cinfo.image_height) {
      row = (JSAMPROW)(((char *)buffers[buf.index].start) + (cinfo.next_scanline * 3 * fmt.fmt.pix.width));
      jpeg_write_scanlines(&cinfo, &row, 1);
   }
   jpeg_finish_compress(&cinfo);
   jpeg_destroy_compress(&cinfo);

   if(!(fout = fopen("output.jpg", "w"))) {
      perror("Cannot open image");
      exit(EXIT_FAILURE);
   }
   fwrite(outbuf, outlen, 1, fout);
   fclose(fout);

   type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
   xioctl(fd, VIDIOC_STREAMOFF, &type);

   for(i = 0; i < req.count; i++) v4l2_munmap(buffers[i].start, buffers[i].length);
   v4l2_close(fd);

   exit(EXIT_SUCCESS);
}
